@@ -2,6 +2,7 @@
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -20,39 +21,50 @@ module Network.HTTP.Client.Request
     , requestBuilder
     , useDefaultTimeout
     , setQueryString
+    , streamFile
+    , observedStreamFile
+    , username
+    , password
     ) where
 
+import Data.Int (Int64)
 import Data.Maybe (fromMaybe, isJust)
 import Data.Monoid (mempty, mappend)
 import Data.String (IsString(..))
 import Data.Char (toLower)
+import Control.Applicative ((<$>))
 import Control.Monad (when, unless)
 import Numeric (showHex)
 
 import Data.Default.Class (Default (def))
 
 import Blaze.ByteString.Builder (Builder, fromByteString, fromLazyByteString, toByteStringIO, flush)
-import Blaze.ByteString.Builder.Char8 (fromChar)
+import Blaze.ByteString.Builder.Char8 (fromChar, fromShow)
 
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Char8 as S8
 import qualified Data.ByteString.Lazy as L
+import Data.ByteString.Lazy.Internal (defaultChunkSize)
 
 import qualified Network.HTTP.Types as W
-import Network.URI (URI (..), URIAuth (..), parseURI, relativeTo, escapeURIString, isAllowedInURI)
+import Network.URI (URI (..), URIAuth (..), parseURI, relativeTo, escapeURIString, isAllowedInURI, isReserved)
 
 import Control.Monad.IO.Class (liftIO)
-import Control.Exception (Exception, toException, throw, throwIO)
+import Control.Exception (Exception, toException, throw, throwIO, IOException)
+import qualified Control.Exception as E
 import qualified Data.CaseInsensitive as CI
 import qualified Data.ByteString.Base64 as B64
 
 import Network.HTTP.Client.Types
+import Network.HTTP.Client.Util
 import Network.HTTP.Client.Connection
 
 import Network.HTTP.Client.Util (readDec, (<>))
-import System.Timeout (timeout)
 import Data.Time.Clock
 import Control.Monad.Catch (MonadThrow, throwM)
+import Data.IORef
+
+import System.IO (withBinaryFile, hTell, hFileSize, Handle, IOMode (ReadMode))
 
 -- | Convert a URL into a 'Request'.
 --
@@ -97,20 +109,44 @@ getUri req = URI
         , uriPort = ':' : show (port req)
         }
     , uriPath = S8.unpack $ path req
-    , uriQuery = S8.unpack $ queryString req
+    , uriQuery =
+        case S8.uncons $ queryString req of
+            Just (c, _) | c /= '?' -> '?' : (S8.unpack $ queryString req)
+            _ -> S8.unpack $ queryString req
     , uriFragment = ""
     }
 
+applyAnyUriBasedAuth :: URI -> Request -> Request
+applyAnyUriBasedAuth uri req =
+    if hasAuth
+      then applyBasicAuth (S8.pack theuser) (S8.pack thepass) req
+      else req
+  where
+    hasAuth = (notEmpty theuser) && (notEmpty thepass)
+    notEmpty = not . null
+    theuser = username authInfo
+    thepass = password authInfo
+    authInfo = maybe "" uriUserInfo $ uriAuthority uri
+
+username :: String -> String
+username = encode . takeWhile (/=':') . authPrefix
+
+password :: String -> String
+password = encode . takeWhile (/='@') . drop 1 . dropWhile (/=':')
+
+encode :: String -> String
+encode = escapeURIString (not . isReserved)
+
+authPrefix :: String -> String
+authPrefix u = if '@' `elem` u then takeWhile (/= '@') u else ""
+
 -- | Validate a 'URI', then add it to the request.
 setUri :: MonadThrow m => Request -> URI -> m Request
 setUri req uri = do
     sec <- parseScheme uri
     auth <- maybe (failUri "URL must be absolute") return $ uriAuthority uri
-    if not . null $ uriUserInfo auth
-        then failUri "URL auth not supported; use applyBasicAuth instead"
-        else return ()
     port' <- parsePort sec auth
-    return req
+    return $ applyAnyUriBasedAuth uri req
         { host = S8.pack $ uriRegName auth
         , port = port'
         , secure = sec
@@ -157,6 +193,7 @@ instance Show Request where
         , "  rawBody              = " ++ show (rawBody x)
         , "  redirectCount        = " ++ show (redirectCount x)
         , "  responseTimeout      = " ++ show (responseTimeout x)
+        , "  requestVersion       = " ++ show (requestVersion x)
         , "}"
         ]
 
@@ -203,6 +240,11 @@ instance Default Request where
                                 then throwIO exc
                                 else return (Just remainingTime, res)
         , cookieJar = Just def
+        , requestVersion = W.http11
+        , onRequestBodyException = \se ->
+            case E.fromException se of
+                Just (_ :: IOException) -> return ()
+                Nothing -> throwIO se
         }
 
 instance IsString Request where
@@ -280,34 +322,69 @@ needsGunzip req hs' =
      && ("content-encoding", "gzip") `elem` hs'
      && decompress req (fromMaybe "" $ lookup "content-type" hs')
 
-requestBuilder :: Request -> Connection -> IO ()
-requestBuilder req Connection {..} =
-    bodySource
+requestBuilder :: Request -> Connection -> IO (Maybe (IO ()))
+requestBuilder req Connection {..}
+    | expectContinue = flushHeaders >> return (Just (checkBadSend sendLater))
+    | otherwise      = sendNow      >> return Nothing
   where
-    writeBuilder = toByteStringIO connectionWrite
+    expectContinue   = Just "100-continue" == lookup "Expect" (requestHeaders req)
+    checkBadSend f   = f `E.catch` onRequestBodyException req
+    writeBuilder     = toByteStringIO connectionWrite
+    writeHeadersWith = writeBuilder . (builder `mappend`)
+    flushHeaders     = writeHeadersWith flush
 
-    (contentLength, bodySource) =
+    (contentLength, sendNow, sendLater) =
         case requestBody req of
-            RequestBodyLBS lbs -> (Just $ L.length lbs, writeBuilder $ builder `mappend` fromLazyByteString lbs)
-            RequestBodyBS bs -> (Just $ fromIntegral $ S.length bs, writeBuilder $ builder `mappend` fromByteString bs)
-            RequestBodyBuilder i b -> (Just $ i, writeBuilder $ builder `mappend` b)
+            RequestBodyLBS lbs ->
+                let body  = fromLazyByteString lbs
+                    now   = checkBadSend $ writeHeadersWith body
+                    later = writeBuilder body
+                in (Just (L.length lbs), now, later)
+
+            RequestBodyBS bs ->
+                let body  = fromByteString bs
+                    now   = checkBadSend $ writeHeadersWith body
+                    later = writeBuilder body
+                in (Just (fromIntegral $ S.length bs), now, later)
+
+            RequestBodyBuilder len body ->
+                let now   = checkBadSend $ writeHeadersWith body
+                    later = writeBuilder body
+                in (Just len, now, later)
 
             -- See https://github.com/snoyberg/http-client/issues/74 for usage
             -- of flush here.
-            RequestBodyStream i stream -> (Just i, writeBuilder (builder `mappend` flush) >> writeStream False stream)
-            RequestBodyStreamChunked stream -> (Nothing, writeBuilder (builder `mappend` flush) >> writeStream True stream)
+            RequestBodyStream len stream ->
+                let body = writeStream False stream
+                    -- Don't check for a bad send on the headers themselves.
+                    -- Ideally, we'd do the same thing for the other request body
+                    -- types, but it would also introduce a performance hit since
+                    -- we couldn't merge request headers and bodies together.
+                    now  = flushHeaders >> checkBadSend body
+                in (Just len, now, body)
+
+            RequestBodyStreamChunked stream ->
+                let body = writeStream True stream
+                    now  = flushHeaders >> checkBadSend body
+                in (Nothing, now, body)
 
     writeStream isChunked withStream =
         withStream loop
       where
         loop stream = do
             bs <- stream
-            when isChunked $
-                connectionWrite $ S8.pack $ showHex (S.length bs) (if S.null bs then "\r\n\r\n" else "\r\n")
-            unless (S.null bs) $ do
-                connectionWrite bs
-                when isChunked $ connectionWrite "\r\n"
-                loop stream
+            if S.null bs
+                then when isChunked $ connectionWrite "0\r\n\r\n"
+                else do
+                    connectionWrite $
+                        if isChunked
+                            then S.concat
+                                [ S8.pack $ showHex (S.length bs) "\r\n"
+                                , bs
+                                , "\r\n"
+                                ]
+                            else bs
+                    loop stream
 
 
     hh
@@ -359,7 +436,13 @@ requestBuilder req Connection {..} =
                     Nothing -> mempty
                     Just ('?', _) -> fromByteString $ queryString req
                     _ -> fromChar '?' <> fromByteString (queryString req))
-            <> fromByteString " HTTP/1.1\r\n"
+            <> (case requestVersion req of
+                    W.HttpVersion 1 1 -> fromByteString " HTTP/1.1\r\n"
+                    W.HttpVersion 1 0 -> fromByteString " HTTP/1.0\r\n"
+                    version ->
+                        fromChar ' ' <>
+                        fromShow version <>
+                        fromByteString "\r\n")
             <> foldr
                 (\a b -> headerPairToBuilder a <> b)
                 (fromByteString "\r\n")
@@ -376,3 +459,40 @@ requestBuilder req Connection {..} =
 -- Since 0.3.6
 setQueryString :: [(S.ByteString, Maybe S.ByteString)] -> Request -> Request
 setQueryString qs req = req { queryString = W.renderQuery True qs }
+
+-- | Send a file as the request body.
+--
+-- It is expected that the file size does not change between calling
+-- `streamFile` and making any requests using this request body.
+--
+-- Since 0.4.9
+streamFile :: FilePath -> IO RequestBody
+streamFile = observedStreamFile (\_ -> return ())
+
+-- | Send a file as the request body, while observing streaming progress via
+-- a `PopObserver`. Observations are made between reading and sending a chunk.
+--
+-- It is expected that the file size does not change between calling
+-- `observedStreamFile` and making any requests using this request body.
+--
+-- Since 0.4.9
+observedStreamFile :: (StreamFileStatus -> IO ()) -> FilePath -> IO RequestBody
+observedStreamFile obs path = do
+    size <- fromIntegral <$> withBinaryFile path ReadMode hFileSize
+
+    let filePopper :: Handle -> Popper
+        filePopper h = do
+            bs <- S.hGetSome h defaultChunkSize
+            currentPosition <- fromIntegral <$> hTell h
+            obs $ StreamFileStatus
+                { fileSize = size
+                , readSoFar = currentPosition
+                , thisChunkSize = S.length bs
+                }
+            return bs
+
+        givesFilePopper :: GivesPopper ()
+        givesFilePopper k = withBinaryFile path ReadMode $ \h -> do
+            k (filePopper h)
+
+    return $ RequestBodyStream size givesFilePopper
